/*============================================================================*/
/*                                                                            */
/*                                                                            */
/*                              boOX 0-227_zenon                              */
/*                                                                            */
/*                      (C) Copyright 2018 Pavel Surynek                      */
/*                                                                            */
/*                http://www.surynek.com | <pavel@surynek.com>                */
/*       http://users.fit.cvut.cz/surynek | <pavel.surynek@fit.cvut.cz>       */
/*                                                                            */
/*============================================================================*/
/* cbsR.h / 0-227_zenon                                                       */
/*----------------------------------------------------------------------------*/
//
// Conflict based search for a semi-continuous version of MAPF.
//
/*----------------------------------------------------------------------------*/


#ifndef __CBSR_H__
#define __CBSR_H__

#include <vector>
#include <list>
#include <set>
#include <map>
#include <unordered_map>
#include <unordered_set>

#include "types.h"
#include "result.h"

#include "core/mapR.h"
#include "core/kruhoR.h"


using namespace std;


/*----------------------------------------------------------------------------*/

namespace boOX
{



    
/*----------------------------------------------------------------------------*/
// sRealCBS

    class sRealCBSBase
    {
    public:
	sRealCBSBase(sRealInstance *real_Instance);
	sRealCBSBase(sRealInstance *real_Instance, sDouble timeout);	

    public:
	sRealInstance *m_real_Instance;
	sDouble m_timeout;
    };

    
/*----------------------------------------------------------------------------*/
// sRealCBS

    class sRealCBS
	: public sRealCBSBase
    {
    public:
	struct Event
	{
	    Event() { /* nothing */ }
	    Event(sInt_32 from_loc_id,
		  sInt_32 to_loc_id,
		  sDouble start_time,
		  sDouble finish_time)
	    : m_from_loc_id(from_loc_id)
	    , m_to_loc_id(to_loc_id)	    
	    , m_start_time(start_time)
	    , m_finish_time(finish_time)		
	    { /* nothing */ }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%s%d --> %d [%.3f, %.3f]\n", indent.c_str(), m_from_loc_id, m_to_loc_id, m_start_time, m_finish_time);
	    }
	    
	    sInt_32 m_from_loc_id, m_to_loc_id;
	    sDouble m_start_time, m_finish_time;
	};

	struct Interval
	{
	    Interval() { /* nothing */ }
	    Interval(sDouble lower, sDouble upper)
	    : m_lower(lower)
	    , m_upper(upper)
	    { /* nothing */ }

	    bool operator<(const Interval &interval) const
	    {
		return (m_upper <= interval.m_lower);
		//return (m_lower < interval.m_lower || (m_lower == interval.m_lower && m_upper < interval.m_upper));
	    }

	    bool operator==(const Interval &interval) const
	    {
		return (m_lower == interval.m_lower && m_upper == interval.m_upper);		
	    }

	    bool overlaps(const Interval &interval) const
	    {
		return (m_upper > interval.m_lower && m_lower < interval.m_upper);
	    }

	    bool empty(void) const
	    {		
		return (m_lower > m_upper);
	    }

	    Interval intersect(const Interval &interval) const
	    {
		return Interval(sMAX(m_lower, interval.m_lower), sMIN(m_upper, interval.m_upper));
	    }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%s[%.3f,%.3f)\n", indent.c_str(), m_lower, m_upper);
	    }	    
	    
	    sDouble m_lower, m_upper;
	};

	
	struct Moment
	{
	    Moment() { /* nothing */ }
	    Moment(sDouble lower, sDouble upper)
	    : m_lower(lower)
	    , m_upper(upper)
	    { /* nothing */ }

	    bool operator<(const Moment &moment) const
	    {
		//return (m_upper <= moment.m_lower);
		return (m_lower < moment.m_lower || (m_lower == moment.m_lower && m_upper < moment.m_upper));
	    }

	    bool operator==(const Moment &moment) const
	    {
		return (m_lower == moment.m_lower && m_upper == moment.m_upper);		
	    }

	    bool overlaps(const Moment &moment) const
	    {
		return (m_upper > moment.m_lower && m_lower < moment.m_upper);
	    }

	    bool empty(void) const
	    {
		return (m_lower > m_upper);
	    }

	    Moment intersect(const Moment &moment) const
	    {
		return Moment(sMAX(m_lower, moment.m_lower), sMIN(m_upper, moment.m_upper));
	    }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%s[%.3f,%.3f)\n", indent.c_str(), m_lower, m_upper);
	    }	    
	    
	    sDouble m_lower, m_upper;
	};	
	
	struct Traversal
	{
	    Traversal() { /* nothing */ }
	    Traversal(sInt_32 kruhobot_id, sInt_32 u_loc_id, sInt_32 v_loc_id, const Interval &interval)
	    : m_kruhobot_id(kruhobot_id)
	    , m_u_loc_id(u_loc_id)
	    , m_v_loc_id(v_loc_id)
	    , m_interval(interval) { /* nothing */ }
	    
	    sInt_32 m_kruhobot_id;
	    sInt_32 m_u_loc_id, m_v_loc_id;
	    Interval m_interval;

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%s%d: %d <-> %d [%.3f,%.3f]\n", indent.c_str(), m_kruhobot_id, m_u_loc_id, m_v_loc_id, m_interval.m_lower, m_interval.m_upper);
	    }
	};

	typedef std::vector<Event> Schedule_vector;
	typedef std::vector<Schedule_vector> KruhobotSchedules_vector;

	struct KruhobotCollision
	{
	    KruhobotCollision() { /* nothing */ }
	    
	    KruhobotCollision(sDouble importance, const Traversal &traversal_A, const Traversal &traversal_B)
	    : m_importance(importance)
	    , m_traversal_A(traversal_A)
	    , m_traversal_B(traversal_B) { /* nothing */ }

	    bool operator<(const KruhobotCollision &kruhobot_collision) const
	    {
		Interval intersection_1 = m_traversal_A.m_interval.intersect(m_traversal_B.m_interval);
		Interval intersection_2 = kruhobot_collision.m_traversal_A.m_interval.intersect(kruhobot_collision.m_traversal_B.m_interval);
		Moment moment_1(intersection_1.m_lower, intersection_1.m_upper);
		Moment moment_2(intersection_2.m_lower, intersection_2.m_upper);		

//		return (intersection_1 < intersection_2);
//		return (m_importance < kruhobot_collision.m_importance);		
		return (moment_1 < moment_2);		
	    }

	    sDouble m_importance;

	    Traversal m_traversal_A;
	    Traversal m_traversal_B;

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%sKruhobot collision [\n", indent.c_str());
		m_traversal_A.to_Stream(fw, indent + s_INDENT);
		m_traversal_B.to_Stream(fw, indent + s_INDENT);		
		fprintf(fw, "%s]\n", indent.c_str());		
	    }	    
	};

	/*
	typedef std::unordered_map<sInt_32, sInt_32> Occupation_umap;
	typedef std::vector<Occupation_umap> Occupations_vector;
	*/
	typedef std::unordered_set<sInt_32> KruhobotIDs_uset;
	
	typedef std::unordered_map<sInt_32, KruhobotIDs_uset> Cooccupation_umap;
	typedef std::vector<Cooccupation_umap> Cooccupations_vector;

	typedef std::multiset<KruhobotCollision, std::less<KruhobotCollision> > KruhobotCollisions_mset;

	struct LocationConflict
	{
	    LocationConflict() { /* nothing */ }
	    LocationConflict(sInt_32 location_id, const Interval &interval, bool infinity = false)
	    : m_infinity(infinity)
	    , m_location_id(location_id)
	    , m_interval(interval) { /* nothing */ }

	    bool operator<(const LocationConflict &location_conflict) const
	    {
		return (m_interval < location_conflict.m_interval);
	    }

	    bool operator==(const LocationConflict &location_conflict) const
	    {
		return (m_interval == location_conflict.m_interval);
	    }

	    bool overlaps(const Interval &interval) const
	    {
		return m_interval.overlaps(interval);
	    }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		if (m_infinity)
		{
		    fprintf(fw, "%s%d+: [%.3f,%.3f]\n", indent.c_str(), m_location_id, m_interval.m_lower, m_interval.m_upper);
		}
		else
		{
		    fprintf(fw, "%s%d: [%.3f,%.3f]\n", indent.c_str(), m_location_id, m_interval.m_lower, m_interval.m_upper);		    
		}
	    }	    

	    bool m_infinity;
	    sInt_32 m_location_id;
	    Interval m_interval;
	};

	struct LinearConflict
	{
	    LinearConflict() { /* nothing */ }
	    LinearConflict(sInt_32 line_u_id, sInt_32 line_v_id, const Interval &interval)
	    : m_line_u_id(line_u_id)
	    , m_line_v_id(line_v_id)	    
	    , m_interval(interval) { /* nothing */ }

	    bool operator<(const LinearConflict &linear_conflict) const
	    {
		return (m_interval < linear_conflict.m_interval);
	    }

	    bool operator==(const LinearConflict &linear_conflict) const
	    {
		return (m_interval == linear_conflict.m_interval);
	    }

	    bool overlaps(const Interval &interval) const
	    {
		return m_interval.overlaps(interval);
	    }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%s%d<->%d: [%.3f,%.3f]\n", indent.c_str(), m_line_u_id, m_line_v_id, m_interval.m_lower, m_interval.m_upper);
	    }

	    sInt_32 m_line_u_id;
	    sInt_32 m_line_v_id;	    
	    Interval m_interval;
	};

	struct Occupation
	{
	    Occupation() { /* nothing */ }
	    Occupation(sInt_32 kruhobot_id, const Interval &interval)
	    : m_kruhobot_id(kruhobot_id)
	    , m_interval(interval) { /* nothing */ }
	    
	    sInt_32 m_kruhobot_id;
	    Interval m_interval;	    
	};

	struct Uline
	{
	    Uline() { /* nothing */ }
	    Uline(sInt_32 u_id, sInt_32 v_id)
	    {
		if (u_id < v_id)
		{
		    m_lower_id = u_id;
		    m_upper_id = v_id;
		}
		else
		{
		    m_lower_id = v_id;
		    m_upper_id = u_id;		    
		}
	    }

	    bool operator<(const Uline &uline) const
	    {
		return (m_lower_id < uline.m_lower_id || (m_lower_id ==  uline.m_lower_id && m_upper_id < uline.m_upper_id));
	    }

	    bool operator==(const Uline &uline) const
	    {
		return (m_lower_id ==  uline.m_lower_id && m_upper_id == uline.m_upper_id);
	    }

	    std::pair<sInt_32, sInt_32> get_Pair(void) const
	    {
		return std::pair<sInt_32, sInt_32>(m_lower_id, m_upper_id);
	    }

	    sInt_32 m_lower_id;
	    sInt_32 m_upper_id;
	};

	typedef std::map<Interval, LocationConflict, std::less<Interval> > LocationConflicts_map;
	typedef std::map<Interval, LinearConflict, std::less<Interval> > LinearConflicts_map;

	typedef std::unordered_map<sInt_32, LocationConflicts_map> LocationConflicts__umap;
	typedef std::map<Uline, LinearConflicts_map, std::less<Uline> > LinearConflicts__map;

	typedef std::vector<LocationConflicts__umap> KruhobotLocationConflicts_vector;
	typedef std::vector<LinearConflicts__map> KruhobotLinearConflicts_vector;	

	struct Transition
	{
	    Transition() { /* nothing */ }
	    Transition(sInt_32 trans_id, sDouble time, sDouble cost, sInt_32 location_id, sInt_32 prev_trans_id)
	    : m_trans_id(trans_id)
	    , m_time(time)
	    , m_cost(cost)
	    , m_location_id(location_id)
	    , m_prev_trans_id(prev_trans_id) { /* nothing */ }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%s%d [%.3f {%.3f}] (%d <-- %d)\n", indent.c_str(), m_location_id, m_time, m_cost, m_trans_id, m_prev_trans_id);
	    }	    	    

	    sInt_32 m_trans_id;
	    
	    sDouble m_time;
	    sDouble m_cost;
	    sInt_32 m_location_id;
	    sInt_32 m_prev_trans_id;	    
	};

	typedef std::unordered_set<sInt_32> LocationIDs_uset;
	typedef std::map<sDouble, LocationIDs_uset> Transitions_map;	
	typedef std::multimap<sDouble, Transition, std::less<sDouble> > Transitions_mmap;
	typedef std::vector<Transition> Transitions_vector;

	typedef std::map<Moment, KruhobotIDs_uset, std::less<Moment>  > Cooccupations_map;	
	typedef std::unordered_map<sInt_32, Cooccupations_map> LocationCooccupations_umap;
	typedef std::map<Uline, Cooccupations_map, std::less<Uline> > LinearCooccupations_map;

	struct Node
	{
	    Node(sInt_32 node_id)
	    : m_node_id(node_id) { /* nothing */ }

	    void to_Screen(const sString &indent = "") const
	    {
		to_Stream(stdout, indent);
	    }
	    
	    void to_Stream(FILE *fw, const sString &indent = "") const
	    {
		fprintf(fw, "%sNode (node_id = %d, upper_id = %d, left_id = %d, right_id = %d, udp_kruhobot_id = %d, {cost = %.3f, make = %.3f, conf = %.3f}) [\n",
			indent.c_str(),
			m_node_id,
			m_upper_node_id,
			m_left_node_id,
			m_right_node_id,
			m_update_kruhobot_id,
			m_cost,
			m_makespan,
			m_conflicting);

		sInt_32 N_kruhobots_1 = m_kruhobot_location_Conflicts.size();
		fprintf(fw, "%s%slocation conflicts: {\n", indent.c_str(), s_INDENT.c_str());
		for (sInt_32 kruhobot_id = 1; kruhobot_id < N_kruhobots_1; ++kruhobot_id)
		{
		    if (!m_kruhobot_location_Conflicts[kruhobot_id].empty())
		    {
			fprintf(fw, "%s%s%d: [\n", indent.c_str(), s2_INDENT.c_str(), kruhobot_id);

			for (LocationConflicts__umap::const_iterator location_Conflicts = m_kruhobot_location_Conflicts[kruhobot_id].begin(); location_Conflicts != m_kruhobot_location_Conflicts[kruhobot_id].end(); ++location_Conflicts)
			{
			    fprintf(fw, "%s%slocation: %d\n", indent.c_str(), s3_INDENT.c_str(), location_Conflicts->first);
			    for (LocationConflicts_map::const_iterator location_conflict = location_Conflicts->second.begin(); location_conflict != location_Conflicts->second.end(); ++location_conflict)
			    {
				location_conflict->second.to_Screen(indent + s4_INDENT);
			    }
			}
			fprintf(fw, "%s%s]\n", indent.c_str(), s2_INDENT.c_str());		    
		    }
		}
		fprintf(fw, "%s%s}\n", indent.c_str(), s_INDENT.c_str());

		N_kruhobots_1 = m_kruhobot_linear_Conflicts.size();		
		fprintf(fw, "%s%slinear conflicts: {\n", indent.c_str(), s_INDENT.c_str());
		for (sInt_32 kruhobot_id = 1; kruhobot_id < N_kruhobots_1; ++kruhobot_id)
		{
		    if (!m_kruhobot_linear_Conflicts[kruhobot_id].empty())
		    {
			fprintf(fw, "%s%s%d: [\n", indent.c_str(), s2_INDENT.c_str(), kruhobot_id);

			for (LinearConflicts__map::const_iterator linear_Conflicts = m_kruhobot_linear_Conflicts[kruhobot_id].begin(); linear_Conflicts != m_kruhobot_linear_Conflicts[kruhobot_id].end(); ++linear_Conflicts)
			{
			    fprintf(fw, "%s%sline: %d<->%d\n", indent.c_str(), s3_INDENT.c_str(), linear_Conflicts->first.m_lower_id, linear_Conflicts->first.m_upper_id);
			    for (LinearConflicts_map::const_iterator linear_conflict = linear_Conflicts->second.begin(); linear_conflict != linear_Conflicts->second.end(); ++linear_conflict)
			    {
				linear_conflict->second.to_Screen(indent + s4_INDENT);
			    }
			}
			fprintf(fw, "%s%s]\n", indent.c_str(), s2_INDENT.c_str());		    
		    }
		}
		fprintf(fw, "%s%s}\n", indent.c_str(), s_INDENT.c_str());

		N_kruhobots_1 = m_kruhobot_Schedules.size();		
		fprintf(fw, "%s%sschedules: {\n", indent.c_str(), s_INDENT.c_str());
		for (sInt_32 kruhobot_id = 1; kruhobot_id < N_kruhobots_1; ++kruhobot_id)
		{
		    fprintf(fw, "%s%s%d: [\n", indent.c_str(), s2_INDENT.c_str(), kruhobot_id);

		    for (Schedule_vector::const_iterator event = m_kruhobot_Schedules[kruhobot_id].begin(); event != m_kruhobot_Schedules[kruhobot_id].end(); ++event)
		    {
			event->to_Screen(indent + s3_INDENT);
		    }
		    fprintf(fw, "%s%s]\n", indent.c_str(), s2_INDENT.c_str());		    
		}
		fprintf(fw, "%s%s}\n", indent.c_str(), s_INDENT.c_str());		

		fprintf(fw, "%s]\n", indent.c_str());
	    }	    

	    sInt_32 m_node_id;

	    sDouble m_cost;
	    sDouble m_makespan;
	    sDouble m_conflicting;

	    sInt_32 m_upper_node_id;
	    sInt_32 m_left_node_id;
	    sInt_32 m_right_node_id;

	    sInt_32 m_update_kruhobot_id;	    

	    KruhobotLocationConflicts_vector m_kruhobot_location_Conflicts;
	    KruhobotLinearConflicts_vector   m_kruhobot_linear_Conflicts;
	    KruhobotSchedules_vector         m_kruhobot_Schedules;
	};

	typedef std::vector<Node> Nodes_vector;

	struct NodeReference
	{
	    NodeReference(sInt_32 node_id, const Nodes_vector *nodes_Store)
	    : m_node_id(node_id)
	    , m_nodes_Store(nodes_Store)
	    {
		// nothing
	    }

	    bool operator<(const NodeReference &node_ref) const
	    {
//		return m_node_id < node_ref.m_node_id;		
//		return ((*m_nodes_Store)[m_node_id].m_makespan < (*m_nodes_Store)[node_ref.m_node_id].m_makespan);	
//		return ((*m_nodes_Store)[m_node_id].m_cost < (*m_nodes_Store)[node_ref.m_node_id].m_cost);
/*
		return ((*m_nodes_Store)[m_node_id].m_cost < (*m_nodes_Store)[node_ref.m_node_id].m_cost
			|| ((*m_nodes_Store)[m_node_id].m_cost == (*m_nodes_Store)[node_ref.m_node_id].m_cost
			    && (*m_nodes_Store)[m_node_id].m_conflicting < (*m_nodes_Store)[node_ref.m_node_id].m_conflicting));
*/
		
		return ((*m_nodes_Store)[m_node_id].m_cost < (*m_nodes_Store)[node_ref.m_node_id].m_cost
			|| ((*m_nodes_Store)[m_node_id].m_cost == (*m_nodes_Store)[node_ref.m_node_id].m_cost
			    && (*m_nodes_Store)[m_node_id].m_makespan < (*m_nodes_Store)[node_ref.m_node_id].m_makespan));
		
/*		
		return ((*m_nodes_Store)[m_node_id].m_makespan < (*m_nodes_Store)[node_ref.m_node_id].m_makespan
			|| ((*m_nodes_Store)[m_node_id].m_makespan == (*m_nodes_Store)[node_ref.m_node_id].m_makespan
			    && (*m_nodes_Store)[m_node_id].m_conflicting < (*m_nodes_Store)[node_ref.m_node_id].m_conflicting));
*/
	    }

	    sInt_32 m_node_id;
	    const Nodes_vector *m_nodes_Store;
	};

	typedef std::multimap<sInt_32, Node, std::less<sInt_32>> Nodes_mmap;

	typedef std::set<Node, std::less<Node> > Nodes_set;
	typedef std::multiset<Node, std::less<Node> > Nodes_mset;

	typedef std::set<NodeReference, std::less<NodeReference> > NodeReferences_set;
	typedef std::multiset<NodeReference, std::less<NodeReference> > NodeReferences_mset;
	typedef std::vector<NodeReference> NodeReferences_vector;

	typedef std::vector<sInt_32> NodeIDs_vector;	
	
    public:
	sRealCBS(sRealInstance *real_Instance);
	sRealCBS(sRealInstance *real_Instance, sDouble timeout);
	/*----------------------------------------------------------------------------*/

	sDouble find_ShortestNonconflictingSchedules(sRealSolution &real_Solution, sDouble cost_limit);
	sDouble find_ShortestNonconflictingSchedules(const  sRealInstance &real_Instance,
						     sRealSolution        &real_Solution,
						     sDouble               cost_limit);
	
	sDouble find_ShortestNonconflictingSchedules(KruhobotSchedules_vector &kruhobot_Schedules, sDouble cost_limit);	
	sDouble find_ShortestNonconflictingSchedules(const sRealInstance      &real_Instance,
						     KruhobotSchedules_vector &kruhobot_Schedules,
						     sDouble                   cost_limit);

	sDouble find_ShortestNonconflictingSchedules(KruhobotSchedules_vector &kruhobot_Schedules, sDouble cost_limit, sDouble extra_cost);
	sDouble find_ShortestNonconflictingSchedules(const sRealInstance      &real_Instance,
						     KruhobotSchedules_vector &kruhobot_Schedules,
						     sDouble                   cost_limit,
						     sDouble                   extra_cost);
	/*----------------------------------------------------------------------------*/	

	sDouble find_NonconflictingSchedules(const sRealInstance              &real_Instance,
					     KruhobotLocationConflicts_vector &kruhobot_location_Conflicts,
					     KruhobotLinearConflicts_vector   &kruhobot_linear_Conflicts,
					     KruhobotSchedules_vector         &kruhobot_Schedules,
					     sDouble                           cost_limit,
					     sDouble                           extra_cost);
	
	sDouble update_NonconflictingSchedule(sInt_32                           upd_kruhobot_id,
					      const sRealInstance              &real_Instance,
					      KruhobotLocationConflicts_vector &kruhobot_location_Conflicts,
					      KruhobotLinearConflicts_vector   &kruhobot_linear_Conflicts,
					      KruhobotSchedules_vector         &kruhobot_Schedules,
					      sDouble                           makespan_bound,
					      sDouble                           cost_limit,
					      sDouble                           extra_cost) const;

	sDouble examine_NonconflictingSchedules(const sRealInstance              &real_Instance,
						sInt_32                           upper_node_id,
						KruhobotLocationConflicts_vector  kruhobot_location_Conflicts,
						KruhobotLinearConflicts_vector    kruhobot_linear_Conflicts,
						KruhobotSchedules_vector          kruhobot_Schedules,
						sDouble                           cost_limit,
						sDouble                           extra_cost,
						Nodes_vector                     &search_Store,
						NodeReferences_mset              &search_Queue) const;

	void introduce_KruhobotConflict(const Traversal                  &kruhobot_traversal,
					KruhobotLocationConflicts_vector &kruhobot_location_Conflicts,
					KruhobotLinearConflicts_vector   &kruhobot_linear_Conflicts,
					bool                              infinity = false) const;	    

	sDouble analyze_NonconflictingSchedules(const sRealInstance            &real_Instance,
						const KruhobotSchedules_vector &kruhobot_Schedules,
						KruhobotCollisions_mset        &kruhobot_Collisions) const;

	sDouble calc_KruhobotCollisionImportance(const Traversal            &traversal_A,
						 const Traversal            &traversal_B,
						 LocationCooccupations_umap &location_Cooccupations,
						 LinearCooccupations_map    &linear_Cooccupations) const;

	sDouble calc_KruhobotCollisionImportance(const Traversal            &traversal,
						 LocationCooccupations_umap &location_Cooccupations,
						 LinearCooccupations_map    &linear_Cooccupations) const;	

	/*----------------------------------------------------------------------------*/
	
	sDouble find_KruhobotNonconflictingSchedule(const sKruhobot               &kruhobot,
						    const s2DMap                  &map,
						    sInt_32                        source_loc_id,
						    sInt_32                        sink_loc_id,
						    sDouble                        cost_limit,
						    sDouble                        extra_cost,
						    const LocationConflicts__umap &location_Conflicts,
						    const LinearConflicts__map    &linear_Conflicts,
						    sDouble                        makespan_bound,						    
						    Schedule_vector               &Schedule) const;

	bool check_KruhobotNonconflictingSchedule(const Schedule_vector         &Schedule,
						  const LocationConflicts__umap &location_Conflicts,
						  const LinearConflicts__map    &linear_Conflicts) const;
	/*----------------------------------------------------------------------------*/

	static void smooth_Schedule(const Schedule_vector &Schedule, Schedule_vector &smooth_Schedule);
	static sDouble augment_Schedules(const sRealInstance &real_Instance, KruhobotSchedules_vector &kruhobot_Schedules);
	static sInt_32 calc_NodeDepth(const Node &node, const Nodes_vector &nodes_Store);

	static sDouble calc_ScheduleMakespan(const sRealInstance &real_Instance, const KruhobotSchedules_vector &kruhobot_Schedules);		
	static sDouble calc_ScheduleCost(const sRealInstance &real_Instance, const KruhobotSchedules_vector &kruhobot_Schedules);
        /*----------------------------------------------------------------------------*/

	static void to_Screen(const Node &node, const Nodes_vector &nodes_Store, const sString &indent = "");
	static void to_Stream(FILE *fw, const Node &node, const Nodes_vector &nodes_Store, const sString &indent = "");

	static void to_Screen(const KruhobotSchedules_vector &kruhobot_Schedules, const sString &indent = "");
	static void to_Stream(FILE *fw, const KruhobotSchedules_vector &kruhobot_Schedules, const sString &indent = "");
    };


/*----------------------------------------------------------------------------*/

} // namespace boOX

#endif /* __CBSR_H__ */
    
